<?php

namespace Drupal\bol\Drush\Commands;

use Consolidation\OutputFormatters\StructuredData\RowsOfFields;
use Drush\Commands\DrushCommands;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a Drush command for listing Drupal site structure.
 */
class BolCommands extends DrushCommands {

  protected $entityTypeManager;
  protected $entityFieldManager;

  public function __construct(EntityTypeManagerInterface $entityTypeManager, EntityFieldManagerInterface $entityFieldManager) {
    $this->entityTypeManager = $entityTypeManager;
    $this->entityFieldManager = $entityFieldManager;
  }

  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager')
    );
  }

  /**
   * Outputs a Bill of Lading of the site's entities and fields.
   *
   * @command bol:report
   * @aliases bol
   * @outputformat table
   * @field-labels
   *   Entity: Entity
   *   Name: Name
   *   ID: ID
   *   Status: Status
   *   Description: Description
   * @default-fields Entity,Name,ID,Status,Description
   *
   * @return \Consolidation\OutputFormatters\StructuredData\RowsOfFields
   */
  public function report() {
    $rows = [];

    $label_overrides = [
      'taxonomy_term' => 'Taxonomy',
      'node' => 'Content',
      'paragraph' => 'Paragraph',
    ];

    foreach ($this->entityTypeManager->getDefinitions() as $entity_type_id => $entity_type) {
      if (!$entity_type->hasKey('id')) {
          continue;
      }
      $label = $entity_type->getLabel();
      $entity_label = $label_overrides[$entity_type_id] ?? $label;

      $bundles = NULL;

      if ($entity_type->hasKey('bundle')) {
        $bundle_entity_type = $entity_type->getBundleEntityType();
        if ($bundle_entity_type && $this->entityTypeManager->hasDefinition($bundle_entity_type)) {
          $bundle_storage = $this->entityTypeManager->getStorage($bundle_entity_type);
          $bundles = $bundle_storage->loadMultiple();

          foreach ($bundles as $bundle_id => $bundle) {
            $rows[] = [
              'Entity' => $entity_label,
              'Name' => $bundle->label(),
              'ID' => "$entity_type_id.$bundle_id",
              'Status' => 'Enabled',
              'Description' => method_exists($bundle, 'getDescription') ? $bundle->getDescription() : '',
            ];

            try {
              if ($entity_type->entityClassImplements('\Drupal\Core\Entity\ContentEntityInterface')) {
                $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle_id);
                foreach ($fields as $field_name => $field_def) {
                  if ($field_def->getFieldStorageDefinition()->isBaseField()) {
                    continue;
                  }
                  $rows[] = [
                    'Entity' => '',
                    'Name' => '↳ ' . $field_def->getLabel(),
                    'ID' => $field_name,
                    'Status' => $field_def->isRequired() ? 'Required' : 'Optional',
                    'Description' => $field_def->getDescription(),
                  ];
                }
              }
            } catch (\Throwable $e) {
              // Skip problematic bundles (e.g., block).
              continue;
            }
          }
        }
      }

      if (empty($bundles)) {
        $rows[] = [
          'Entity' => $entity_label,
          'Name' => $entity_label,
          'ID' => $entity_type_id,
          'Status' => 'Enabled',
          'Description' => method_exists($entity_type, 'getDescription') ? $entity_type->getDescription() : '',
        ];

        try {
          if ($entity_type->entityClassImplements('\Drupal\Core\Entity\ContentEntityInterface')) {
            $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, NULL);
            foreach ($fields as $field_name => $field_def) {
              if ($field_def->getFieldStorageDefinition()->isBaseField()) {
                continue;
              }
              $rows[] = [
                'Entity' => '',
                'Name' => '↳ ' . $field_def->getLabel(),
                'ID' => $field_name,
                'Status' => $field_def->isRequired() ? 'Required' : 'Optional',
                'Description' => $field_def->getDescription(),
              ];
            }
          }
        } catch (\Throwable $e) {
          // Ignore entity types that throw on field inspection.
        }
      }
    }

    return new RowsOfFields($rows);
  }
}
